﻿{-# LANGUAGE DeriveDataTypeable #-}
{-# LANGUAGE TupleSections #-}

import Control.Monad
import Control.Monad.Trans
import Control.Monad.State (StateT, execStateT, evalStateT, runStateT)
import qualified Control.Monad.State as ST (get, gets, modify)
import Control.Applicative

import Data.Typeable

import Data.List
import Data.Maybe

import Prelude hiding (read)

import PView

-------------------------------------------------------------------------------

data Void

 -- monad transformer for accessing ship data
type ShipsT     = StateT ShipsState IO
type ShipsState = ([Ship], [Position])
type ShipsView q a b = PView q ShipsT a b

type Coord = (Double, Double)

data Ship = Ship
  {  s_name         :: String
  ,  s_capacity     :: Int
  } deriving Show

data Position = Position
  {  p_ship_name    :: String
  ,  p_position     :: Coord
  ,  p_lastreading  :: Int  
  } deriving Show
  
baseData = (
   [Ship {s_name = "Queen", s_capacity = 100},
    Ship {s_name = "Aberford", s_capacity = 200},
    Ship {s_name = "Ross", s_capacity = 80}], 
   [Position {p_ship_name = "Ross", p_position = (10,25), p_lastreading = 1000},
    Position {p_ship_name = "Queen", p_position = (100,200), p_lastreading = 1050},
	Position {p_ship_name = "Aberford", p_position = (50,50), p_lastreading = 900}])
   
-- data source definitions of ships and their positions   
data ShipSource = ShipSource deriving (Typeable, Eq)
data PositionSource = PositionSource deriving (Typeable, Eq)

ships :: ShipsView ShipSource [Ship] [Ship]
ships = Source (MkSource (const $ ST.gets fst)
                 (const $ \ss -> ST.modify (\(_, ps) -> (ss, ps)) >> return (const True)))

positions :: PView PositionSource ShipsT [Position] [Position]
positions = Source (MkSource (const $ ST.gets snd)
                 (const $ \ps -> ST.modify (\(ss, _) -> (ss, ps)) >> return (const True)))

-- view definitions				 				 
data Q1 = Q1 {unQ1 :: Int} deriving Typeable

q1' :: ShipsView (ShipSource,Q1) [Ship] [Ship]
q1' = ships `Focus` (MkPLens {getf = get, putf = put}) where
  
  pred (Q1 c) s = s_capacity s >= c
     
  get = filter . pred
  put p rs ws = (m' ++ ws, 
    \q -> unQ1 q >= minimum (map s_capacity m ++ map s_capacity ws)) 
    where
       (m, m') = partition (pred p) rs

q1 :: ShipsView Q1 [Ship] [Ship]
q1 = Translate q1' (\q -> (ShipSource,q))

data Q2a = Q2a {unQ2a :: String} deriving Typeable

q2a' :: ShipsView (PositionSource, Q2a) [Position] [Position]
q2a' = positions `Focus` MkPLens {getf = get, putf = put}
   where
     pred (Q2a n) s = p_ship_name s == n
     get = filter . pred
     put q@(Q2a n) rs ws = (filter (not . pred q) rs ++ ws, \q -> unQ2a q == n)

-- a view of those ships which positions are not updated since a given time
-- not Merged with the other views
data Q2b = Q2b Int deriving Typeable

q2b :: ShipsView (PositionSource, Q2b) [String] Void
q2b = split `Project` (MkLens {get = fmap p_ship_name, put = undefined})
   where
     split = positions `Focus` (MkPLens {getf = getf, putf = undefined})
     getf (Q2b n) ps = filter ((< n) . p_lastreading) ps     

overlap ((ax1,ay1),(ax2,ay2)) ((bx1,by1),(bx2,by2)) 
  = ay1 >= by1 && ay1 <= by2 && ax2 >= bx1 && ax1 <= bx2         
	 
data Q3 = Q3 Int (Coord, Coord) deriving Typeable

q3' :: ShipsView ((ShipSource,PositionSource),Q3) [(Ship, Position)] [(Ship, Position)]
q3' = joinV `Focus` filterL
  where	
	productV = ships `Product` positions
	joinV = productV `Project` joinL
	joinL = MkLens {get = get, put = const unzip}
	get (ss,ps) = mapMaybe f ss 
	  where f s = (s,) <$> find ((s_name s ==) . p_ship_name) ps
     	 
	filterL = MkPLens {getf = filter . pred, putf = put}
	put q@(Q3 _ r) rs ws = (m' ++ ws, inval)
	  where
	    (m, m') = partition (pred q) rs
	    inval (Q3 c r') = c >= minimum (map (s_capacity.fst) (m ++ ws)) && overlap r r'

	pred (Q3 c ((x,y),(x',y'))) (s,p) = s_capacity s >= c && inside (p_position p)
	  where
	    inside (a,b) = a >= x && a <= x' && b >= y && b <= y'
		
q2a :: ShipsView Q2a (Maybe Coord) (Coord, Int)
q2a = Translate pr (Left . (PositionSource,))
  where
    pr = Project q' (MkLens {get = get, put = put})
    get ss = fmap p_position (listToMaybe (map snd ss))
    put ss (c,l) = map (\(_,s) -> (Nothing, s {p_position = c, p_lastreading = l})) ss   
	
q3 :: ShipsView Q3 [String] Void
q3 = Translate pr (Right . ((ShipSource, PositionSource),))
  where
    pr = Project q' (MkLens {get = get, put = undefined})
    get = map (p_ship_name . snd)
    
q' :: ShipsView (Either (PositionSource, Q2a) ((ShipSource,PositionSource),Q3)) [(Maybe Ship, Position)] [(Maybe Ship, Position)]
q' = Merge lv rv invalQ2Q3 invalQ3Q2
  where
    lv = Project q2a' (MkLens {get = map (Nothing,) , put = const (map snd)})
    rv = Project q3' (MkLens {get = map (\(s,p) -> (Just s, p)), put = const (map (\(s, p) -> (fromJust s, p)))})
    
    invalQ2Q3 _ rs ws (_,Q3 _ ((x1,y1),(x2,y2))) = (not.null) (filter (inside.p_position) s2)
      where 
       (_, s2) = unzip (rs++ws)
       inside (x,y) = x >= x1 && x <= x2 && y >= y1 && y <= y2
	   
    invalQ3Q2 _ rs ws (_,Q2a ship) = elem ship $ fmap s_name (catMaybes s1) ++ fmap p_ship_name s2
      where 
       (s1, s2) = unzip (rs++ws)

	   
main = runStateT (evalStateT mainShip []) baseData >>= print

-- list ships which have cargo capacity at least 90
--mainShip = read q1 (Q1 90)

-- read the current coordinates of "Ross"
--mainShip = read q2a (Q2a "Ross")

-- update the current coordinates of "Ross"
--mainShip = update q2a (Q2a "Ross") ((3,3),12) -- >> read q2a (Q2a "Ross")

-- list the name of ships which have a cargo capacity of at elast 50 and located in the given area
--mainShip = read q3 (Q3 50 ((0,0),(100,100)))

mainShip = observe q3 (Q3 80 ((0,0),(20,20))) "oid1" (lift $ print "match") >> update q2a (Q2a "Ross") ((11,12),1200)
